#ifndef WASM_RELOC
#error "WASM_RELOC must be defined"
#endif

WASM_RELOC(R_WASM_FUNCTION_INDEX_LEB,    0)
WASM_RELOC(R_WASM_TABLE_INDEX_SLEB,      1)
WASM_RELOC(R_WASM_TABLE_INDEX_I32,       2)
WASM_RELOC(R_WASM_MEMORY_ADDR_LEB,       3)
WASM_RELOC(R_WASM_MEMORY_ADDR_SLEB,      4)
WASM_RELOC(R_WASM_MEMORY_ADDR_I32,       5)
WASM_RELOC(R_WASM_TYPE_INDEX_LEB,        6)
WASM_RELOC(R_WASM_GLOBAL_INDEX_LEB,      7)
WASM_RELOC(R_WASM_FUNCTION_OFFSET_I32,   8)
WASM_RELOC(R_WASM_SECTION_OFFSET_I32,    9)
WASM_RELOC(R_WASM_EVENT_INDEX_LEB,      10)
WASM_RELOC(R_WASM_MEMORY_ADDR_REL_SLEB, 11)
WASM_RELOC(R_WASM_TABLE_INDEX_REL_SLEB, 12)
