export function print() {
  console.log();
}
export function print_i32(arg) {
  console.log(arg, ': i32');
}
export function print_f32(arg) {
  console.log(arg, ': f32');
}
export function print_f64(arg) {
  console.log(arg, ': f64');
}
export function print_i32_f32(arg0, arg1) {
  console.log(arg0, ': i32');
  console.log(arg1, ': f32');
}
export function print_f64_f64(arg0, arg1) {
  console.log(arg0, ': f64');
  console.log(arg1, ': f64');
}
